local addonName = "DiminiBar"
local chatShort = "drbar"
local chatLong = addonName
local addonDbName = addonName .. 'DB'
local addon = LibStub('AceAddon-3.0'):NewAddon(addonName, 'AceConsole-3.0', 'AceEvent-3.0', 'LibDebugLog-1.0', 'LibBars-1.0', 'AceTimer-3.0')
--local ldb   = LibStub:GetLibrary('LibDataBroker-1.1')
local L = LibStub('AceLocale-3.0'):GetLocale(addonName)
--local bars = LibStub('LibBars-1.0')
local _G = _G
_G[addonName] = addon

local options = {
	name = addonName,
	handler = addon,
	type = 'group',
	args = {
 		enable = {
 			name = "Enable",
 			desc = "Enable addon.",
 			type = 'toggle',
 			get = 'getEnable',
 			set = 'setEnable',
 		},
		layout = {
			name = "Layout",
			desc = "Layout options.",
			type = 'group',
			args = {
				anchor = {
					name = "Show anchor",
					desc = "Show anchor. This makes the bars movable.",
					type = 'toggle',
					get = 'getUnlock',
					set = 'setUnlock',
				},
				x = {
					name = "X Position",
					desc = "Horizonzal position of the bars.",
					type = 'range',
					min = 0,
					max = 1000,
					step = 1,
					get = 'getX',
					set = 'setX',
				},
				y = {
					name = "Y Position",
					desc = "Vertical position of the bars.",
					type = 'range',
					min = 0,
					max = 1000,
					step = 1,
					get = 'getY',
					set = 'setY',
				},
				width = {
					name = "Width",
					desc = "Width of the bars.",
					type = 'range',
					min = 10,
					max = 1000,
					step = 1,
					get = 'getWidth',
					set = 'setWidth',
				},
				height = {
					name = "Height",
					desc = "Height of one bar.",
					type = 'range',
					min = 5,
					max = 50,
					step = 1,
					get = 'getHeight',
					set = 'setHeight',
				},
				fontSize = {
					name = "Font size",
					desc = "Font size of the text.",
					type = 'range',
					min = 6,
					max = 48,
					step = 1,
					get = 'getFontSize',
					set = 'setFontSize',
				},
			}, -- layout.args
		}, -- layout

		
	},
}


-- The CC Data
local polySap = "POLY/SAP"
local poly = {category = polySap, duration = 10, pve=false, pvp=true }
local fearBlind = "FEAR/BLIND"
local charge = "CHARGE"
local stun = "STUN"
local cheapShot = "CHEAP/POUNCE"
local freezeRoot = "ROOT"

addon.ccTable = {
	[61305] = poly,		-- cat
	[28271] = poly,		-- turtle
	[12826] = poly,		-- sheep
	[28272] = poly,		-- pig
	[61721] = poly,		-- rabbit
	[10890] = {category = fearBlind, 	duration = 8,	pve=false,	pvp=true},		-- psychic scream
	[7922]  = {category = charge, 		duration = 1.5,	pve=true,	pvp=false}, 	-- charge stun
	[20253] = {category = stun, 		duration = 3, 	pve=true,	pvp=true}, 	-- intercept
	[10308] = {category = stun,   		duration = 6, 	pve=true,	pvp=true}, 		-- hammer of justice
	[20066] = {category = polySap,		duration = 6,	pve=false,	pvp=true},	 	-- repentance
	[46968] = {category = stun,			duration = 4,	pve=true,	pvp=true},		-- shockwave
	[12809] = {category = stun,			duration = 5, 	pve=true,	pvp=true},		-- concussing blow
	[408] 	= {category = stun,			duration = 5, 	pve=true,	pvp=true},		-- kidney shot 5
	[2070]  = {category = polySap,  	duration = 10,	pvp=true,	pve=false},		-- sap
	[1833]  = {category = cheapShot,	duration = 4, 	pve=true,	pvp=true},		-- cheap shot 
	[122]	= {category = freezeRoot, 	duration = 8,	pvp=true,	pve=false},		-- frost nova
	-- TODO Deep Freeze
}

local BREAK = {}
local APPLY = {}

--[[
-- Getter Setter block.
--]]
function addon:getEnable(info)
	return self.db.profile.enabled
end

function addon:setEnable(info,value)
	self.db.profile.enabled = value
	if value then
		-- enable
		self:CreateBars()
	else
		-- disable
		if self.oOMBar then
			self.oOMBar:Hide()
		end
	end
end

-- function addon:getOOMBar(info)
-- 	return self.db.profile.showBar
-- end

-- function addon:setOOMBar(info,value)
-- 	self.db.profile.showBar = value
-- 	if value then
-- 		-- show
-- 		self:CreateBars()
-- 	else
-- 		-- hide
--		if self.oOMBar then
--			self.oOMBar:Hide()
--		end
-- 	end
-- end

function addon:getUnlock(info)
	return self.db.profile.unlock
end

function addon:setUnlock(info,value)
	self.db.profile.unlock = value
	if value and self.bars then
		-- show
		self.bars:ShowAnchor()
	elseif self.bars then
		-- hide
		self.bars:HideAnchor()
	end
end

-- function addon:getMinPercent()
-- 	return self.db.profile.minPercent
-- end

-- function addon:setMinPercent(info,value)
-- 	self.db.profile.minPercent = value
-- end

-- function addon:getTheta()
-- 	return self.db.profile.theta
-- end

-- function addon:setTheta(info,value)
-- 	self.db.profile.theta = value
-- end

-- function addon:getTimeWindow()
-- 	return self.db.profile.timeWindow
-- end

-- function addon:setTimeWindow(info,value)
-- 	self.db.profile.timeWindow = value
-- end

-- function addon:getNumCasts()
-- 	return self.db.profile.numCasts
-- end

-- function addon:setNumCasts(info,value)
-- 	self.db.profile.numCasts = value
-- end


-- function addon:getMaxTime()
-- 	return self.db.profile.maxTime
-- end

-- function addon:getAlgorithm() 
-- 	return self.db.profile.algorithm
-- end

-- function addon:setAlgorithm(info,value)
-- 	if value == 'exp' or value == 'avg' or value == 'cast' then
-- 		self.db.profile.algorithm = value
-- 		self:ResetCalc()
-- 	end
-- end

-- function addon:setMaxTime(info,value)
-- 	self.db.profile.maxTime = value
-- end

-- function addon:getTreshold()
-- 	return self.db.profile.regPercentTreshold
-- end

-- function addon:setTreshold(info,value)
-- 	self.db.profile.regPercentTreshold = value
-- end

function addon:getX()
	return self.db.profile.xPos
end

function addon:setX(info,value)
	self.db.profile.xPos = value
	if self.bars then
		self.bars:SetPoint('TOPLEFT', UIParent, 'BOTTOMLEFT', self.db.profile.xPos, self.db.profile.yPos)
	end
end

function addon:getY()
	return self.db.profile.yPos
end

function addon:setY(info,value)
	self.db.profile.yPos = value
	if self.bars then
		self.bars:SetPoint('TOPLEFT', UIParent, 'BOTTOMLEFT', self.db.profile.xPos, self.db.profile.yPos)
	end
end

function addon:getWidth()
	return self.db.profile.width
end

function addon:setWidth(info,value)
	self.db.profile.width = value
	if self.bars then
		self.bars:SetWidth(value)
	end
end

function addon:getHeight()
	return self.db.profile.height
end

function addon:setHeight(info,value)
	self.db.profile.height= value
	if self.bars then
		self.bars:SetHeight(value)
	end
end

function addon:getFontSize()
	return self.db.profile.fontSize
end

function addon:setFontSize(info,value)
	self.db.profile.fontSize = value
	if self.bars then
		self.bars:SetFont(nil,value,nil)
	end
	if self.bars then
		for guid,data in self.bars do
			for catId, cat in data do
				cat:SetFont(nil,value,nil)
			end
		end
	end
end


local defaults = {
	profile = {
		unlock = false,
		enabled = true,
-- 		minPercent = 0.02,
-- 		theta = 0.1,
-- 		regPercentTreshold = 0.1,
		showBar = true,
-- 		maxTime = 120,
-- 		algorithm = 'avg',
-- 		timeWindow = 20,
-- 		numCasts = 3,
-- 		width = 150,
		height = 12,
		fontSize = 8,
		xPos = 300,
		yPos = 300,
	},
}

--[[
-- Handles chat commands.
--]]
function addon:ChatCommand(input)
    if not input or input:trim() == '' then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
    else
        LibStub('AceConfigCmd-3.0').HandleCommand(addonName, chatShort, addonName, input)
    end
end

--[[
--	Called when the addon is loaded
--]]
function addon:OnInitialize()
	self.db = LibStub('AceDB-3.0'):New(addonDbName, defaults)
	LibStub('AceConfig-3.0'):RegisterOptionsTable(addonName, options)
	self.optionsFrame = LibStub('AceConfigDialog-3.0'):AddToBlizOptions(addonName, addonName)
	self:RegisterChatCommand(chatShort, 'ChatCommand')
	self:RegisterChatCommand(chatLong, 'ChatCommand')
	do
		local player, _ = UnitName('player')
		if player == 'Chiloe' then
			self:ToggleDebugLog(true)
		end
	end
end

--[[
--	Called when the addon is enabled.
--]]
function addon:OnEnable()
	--self:RegisterEvent('UNIT_SPELLCAST_FAILED')
	--self:RegisterEvent('UNIT_SPELLCAST_FAILED_QUIET')
	--self:RegisterEvent('UNIT_SPELLCAST_SUCCEEDED')
	self:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	
	self:ScheduleRepeatingTimer('CleanUp',60)
	self:CreateBars()
end

function addon:CreateBars()
	if self.db.profile.enabled and not self.bars then
		self.bars = self:NewBarGroup(addonName, nil, self.db.profile.width, self.db.profile.height)
		self.bars:SetFont(nil, self.db.profile.fontSize)
-- 		self.bars:SetColorAt(1.00, 1, 0, 0, 1)
-- 		self.bars:SetColorAt(0.66, 1, 1, 0, 1)
-- 		self.bars:SetColorAt(0.33, 0, 1, 1, 1)
-- 		self.bars:SetColorAt(0.00, 0, 0, 1, 1)
		self.bars:SetPoint('TOPLEFT', UIParent, 'BOTTOMLEFT', self.db.profile.xPos, self.db.profile.yPos)
--		self.bars:SetScale(0.9)
		self.bars.RegisterCallback(self, "AnchorMoved")
		--self.oOMBar = self.bars:NewTimerBar("TTOOM",nil,self.db.profile.maxTime,self.db.profile.maxTime,true)		
		--self.oOMBar:Hide()
		self.bars:HideIcon()
	end
	if self.bars then
		if self.db.profile.unlock then
			self.bars:ShowAnchor()
		else
			self.bars:HideAnchor()
		end
	end
end

function addon:OnDisable()
	-- TODO
end

function addon:AnchorMoved(event,id,x,y)
	self.db.profile.xPos = x
	self.db.profile.yPos = y
end

--[[
--  Updates the displayed bar (if enabled)
--
--]]
function addon:UpdateBar(guid,name,category,maxTime,spellName)
	if not self.db.profile.showBar then return end
		
	--self:Debug("Update")
	if not (self.ccData and self.ccData[guid] and self.ccData[guid][category]) then		
		return
	end
	local data = self.ccData[guid][category]
	local stacks = data.count
	local value, text
	if data.state == APPLY then
		
		if stacks == 1 then
			value = maxTime
		elseif stacks == 2 then
			value = maxTime/2
		elseif stacks == 3 then
			value = maxTime/4
		else
			return
		end
		text = name .. " - " .. category .. ": " .. spellName .. "(" .. stacks .. ")"
	else
		-- BREAK
		value = 15
		if stacks < 3 then
			text = name .. " - " .. category .. ": " .. "DR (" .. stacks .. ")"
		else 
			text = name .. " - " .. category .. ": " .. "Immune"
		end
	end

	self.bar = self.bar or {}
	self.bar[guid] = self.bar[guid] or {}
	
	local barName = addonName .. '.' .. guid .. '.' .. category
	
	if (not self.bar[guid][category]) or (not self.bars:GetBar(barName)) then
		self.bar[guid][category] = self.bars:NewTimerBar(barName,text,value,value,nil,0)
	else
		self.bar[guid][category]:SetText(text)
		self.bar[guid][category]:SetTimer(value,value)
	end	
	self.bar[guid][category]:Show()
end

function addon:UpdateTimer(guid,name,data,action,spellName)
	local category = data.category
	--self:Debug("UpdateTimer",data,data.category)
	-- Ensure Tables exist
	self.ccData = self.ccData or {}
	local ccData = self.ccData
	ccData[guid] = ccData[guid] or {}
	local guidData = ccData[guid]
	guidData[category] = guidData[category] or {}
	local categoryData = guidData[category]
	
	local time = GetTime()
	-- Cleanup
	if categoryData.time and categoryData.time + 15 < time and categoryData.state == BREAK then
		categoryData = {}
		guidData[category] = categoryData
	end
	
	local oldTime = categoryData.time
	categoryData.time = time
	categoryData.count = categoryData.count or 0
	if action == APPLY then
		categoryData.count = categoryData.count + 1
	end
	categoryData.state = action
	
	self:UpdateBar(guid,name,category,data.duration,spellName)
	
	if action==APPLY then
		self:Debug(category,name,"apply",categoryData.count)
	else
		if oldTime then
			self:Debug(category,name,"break",categoryData.count,time-oldTime)
		else
			self:Debug(category,name,"break",categoryData.count,"noTime")
		end
	end
end

--- Checks for ancient data and cleans it.
function addon:CleanUp()
	if not self.ccData then return end
	self:Debug("Enter Clean")
	local time = GetTime()
	for guid,guidData in pairs(self.ccData) do
		for ccId,ccData in pairs(guidData) do
			-- large value to prevent too early cleaning
			if ccData.time and ccData.time + 60 < time then
				self:Debug("Purging",ccId,"from",guid)
				guidData[ccId] = nil
			end
		end
		if (# guidData) == 0 then
			self:Debug("Purging",guid)
			self.ccData[guid] = nil
		end
	end
	-- Todo: Bars
	for guid,guidData in pairs(self.bar) do
		for ccId,ccData in pairs(guidData) do
			-- large value to prevent too early cleaning
			local barName = addonName .. '.' .. guid .. '.' .. ccId
		
			if not self.bars:GetBar(barName) then			
				self:Debug("Purging bar",ccId,"from",guid)
				guidData[ccId] = nil
			end
		end
		if (# guidData) == 0 then
			self:Debug("Purging bars ",guid)
			self.bar[guid] = nil
		end
	end
end


-- function addon:UNIT_SPELLCAST_SUCCEEDED(event,unit,name,rank)
-- 	if not self.db.profile.enabled then return end
-- 	-- Maybe TODO
-- end

function addon:COMBAT_LOG_EVENT_UNFILTERED(event,timestamp,realEvent,...)
	if realEvent=='SPELL_AURA_REMOVED' then
		self:SPELL_AURA_REMOVED(realEvent,...)
	elseif realEvent=='SPELL_AURA_APPLIED' then
		self:SPELL_AURA_APPLIED(realEvent,...)
	elseif realEvent=='SPELL_AURA_REFRESH' then
		self:SPELL_AURA_APPLIED(realEvent,...)
	end
end

local ignoreSpells = {
	[57073] = true,
 	[1715]  = true, -- harmstring
	[46857] = true,
	[12721] = true,
	[2457]  = true,
	[2458]  = true,
	[65156] = true,
	[29156] = true,
	[53601] = true,
	[65014] = true,
	[20053] = true,
	[59620] = true,
	[54203] = true,
	[54428] = true,
	[8632]	= true,
	[8640]  = true,	
	[17364] = true,
	[49231] = true,
	[18118] = true,
	[47811] = true,
	[63311] = true,
	[61291] = true,
	[55095] = true,
	[55078] = true,
	[47865] = true,
	[51585] = true,	-- Blade Twisting
	[57970] = true,	-- Deadly Poison IX
	[57975] = true,	-- Wound Poison
	[58683] = true,	-- Savage Combat
	[47855] = true,	-- Drain Soul
}

local function debugSpam(event,sourceName,destName,spellId,spellName)
	if ignoreSpells[spellId] then 
		return 
	end
	addon:Debug(event,sourceName,destName,spellId,spellName)
end

function addon:SPELL_AURA_REMOVED(event,sourceGuid,sourceName,sourceFlags,destGuid,destName,destFlags,spellId,spellName,...)
	if not self.db.profile.enabled then return end
	
	-- not hostile and not neutral?
	if bit.band(destFlags, COMBATLOG_OBJECT_REACTION_MASK) ~= COMBATLOG_OBJECT_REACTION_HOSTILE 
	and bit.band(destFlags, COMBATLOG_OBJECT_REACTION_MASK) ~= COMBATLOG_OBJECT_REACTION_NEUTRAL then
		return
	end
	
	-- get the unit type
	local isPlayer = bit.band(destFlags, COMBATLOG_OBJECT_TYPE_MASK) == COMBATLOG_OBJECT_TYPE_PLAYER
	-- TODO Discover if the unit is a special one which behaves like pvp.
	-- For example Faction Champions ToCr
		
	if self.ccTable[spellId] then
		local data = self.ccTable[spellId]		
		--self:Debug(event,data,data.category)			
		self:UpdateTimer(destGuid,destName,data,BREAK,spellName)
	else
		debugSpam(event,sourceName,destName,spellId,spellName)
	end
		
end

function addon:SPELL_AURA_APPLIED(event,sourceGuid,sourceName,sourceFlags,destGuid,destName,destFlags,spellId,spellName,...)
	if not self.db.profile.enabled then return end
	
	-- not hostile and not neutral?
	if bit.band(destFlags, COMBATLOG_OBJECT_REACTION_MASK) ~= COMBATLOG_OBJECT_REACTION_HOSTILE 
	and bit.band(destFlags, COMBATLOG_OBJECT_REACTION_MASK) ~= COMBATLOG_OBJECT_REACTION_NEUTRAL then
		return
	end
	
	-- get the unit type
	local isPlayer = bit.band(destFlags, COMBATLOG_OBJECT_TYPE_MASK) == COMBATLOG_OBJECT_TYPE_PLAYER
	-- TODO Discover if the unit is a special one which behaves like pvp.
	-- For example Faction Champions ToCr
	
	if self.ccTable[spellId] then
		local data = self.ccTable[spellId]
		--self:Debug(event,data,data.category)
		self:UpdateTimer(destGuid,destName,data,APPLY,spellName)
	else
		debugSpam(event,sourceName,destName,spellId,spellName)
	end
		
end

-- function addon:UNIT_SPELLCAST_FAILED_QUIET(event,unit,name,rank)
-- 	if not self.db.profile.enabled then return end
-- 	self:Debug(event,unit,name,rank)
-- end

-- function addon:UNIT_SPELLCAST_FAILED(event,unit,name,rank)
-- 	if not self.db.profile.enabled then return end
-- 	self:Debug(event,unit,name,rank)
-- end
